from .. import sanitize_msg_param

BYTE_STUFFING_VALUE = b'\xAA'

ServiceNames = {
    # Diagnostic and Communications Management
    0x10 : "Diagnostic Session Control",
    0x11 : "ECU Reset",
    0x27 : "Security Access",
    0x28 : "Communication Control",
    0x29 : "Authentication",
    0x3E : "Tester Present",
    0x83 : "Access Timing Parameters",
    0x84 : "Secured Data Transmission",
    0x85 : "Control DTC Settings",
    0x86 : "Response On Event",
    0x87 : "Link Control",
    # Data Transmission
    0x22 : "Read Data By Identifier",
    0x23 : "Read Memory By Address",
    0x24 : "Read Scaling Data By Identifier",
    0x2A : "Read Data By Identifier Periodic",
    0x2C : "Dynamically Define Data Identifier",
    0x2E : "Write Data By Identifier",
    0x3D : "Write Memory By Address",
    # Stored Data Transmission
    0x14 : "Clear Diagnostic Information",
    0x19 : "Read DTC Information",
    # Input / Output Control
    0x2F : "Input Output Control By Identifier",
    # Remote Activation of Routine
    0x31 : "Routine Control",
    # Upload / Download
    0x34 : "Request Download",
    0x35 : "Request Upload",
    0x36 : "Transfer Data",
    0x37 : "Request Transfer Exit",
    0x38 : "Request File Transfer",
    # Negative Response (not a service)
    0x3F : "Negative Response"
}
"""Dict of service names, organized by UDS Request SID."""

ResponseCodes = {
    0x00 : "Positive Response",
    0x10 : "General Reject",
    0x11 : "Service Not Supported",
    0x12 : "Sub-Function Not Supported",
    0x13 : "Incorrect Message Length Or Invalid Format",
    0x14 : "Response Too Long",
    0x21 : "Busy - Repeat Request",
    0x22 : "Conditions Not Correct",
    0x24 : "Request Sequence Error",
    0x25 : "No Response From Subnet Component",
    0x26 : "Failure Prevents Execution Of Requested Action",
    0x31 : "Request Out Of Range",
    0x33 : "Security Access Denied",
    0x34 : "Authentication Required",
    0x35 : "Invalid Key",
    0x36 : "Exceed Number Of Attempts",
    0x37 : "Required Time Delay Not Expired",
    0x38 : "Secure Data Transmission Required",
    0x39 : "Secure Data Transmission Not Allowed",
    0x3A : "Secure Data Verification Failed",
    0x50 : "Certificate verification failed - Invalid Time Period",
    0x51 : "Certificate verification failed - Invalid Signature",
    0x52 : "Certificate verification failed - Invalid Chain of Trust",
    0x53 : "Certificate verification failed - Invalid Type",
    0x54 : "Certificate verification failed - Invalid Format",
    0x55 : "Certificate verification failed - Invalid Content",
    0x56 : "Certificate verification failed - Invalid Scope",
    0x57 : "Certificate verification failed - Invalid Certificate (revoked)",
    0x58 : "Ownership verification failed",
    0x59 : "Challenge calculation failed",
    0x5A : "Setting Access Rights failed",
    0x5B : "Session key creation/derivation failed",
    0x5C : "Configuration data usage failed",
    0x5D : "De-Authentication failed",
    0x70 : "Upload/Download Not Accepted",
    0x71 : "Transfer Data Suspended",
    0x72 : "General Programming Failure",
    0x73 : "Wrong Block Sequence Counter",
    0x78 : "Request Correctly Received - Response Pending",
    0x7E : "Sub-Function Not Supported In Active Session",
    0x7F : "Service Not Supported In Active Session",
    0x81 : "RPM Too High",
    0x82 : "RPM Too Low",
    0x83 : "Engine Is Running",
    0x84 : "Engine Is Not Running",
    0x85 : "Engine Run Time Too Low",
    0x86 : "Temperature Too High",
    0x87 : "Temperature Too Low",
    0x88 : "Vehicle Speed Too High",
    0x89 : "Vehicle Speed Too Low",
    0x8A : "Throttle/Pedal Too High",
    0x8B : "Throttle/Pedal Too Low",
    0x8C : "Transmission Range Not In Neutral",
    0x8D : "Transmission Range Not In Gear",
    0x8F : "Brake Switch(es) Not Closed (Brake Pedal not pressed or not applied)",
    0x90 : "Shifter Lever Not In Park",
    0x91 : "Torque Converter Clutch Locked",
    0x92 : "Voltage Too High",
    0x93 : "Voltage Too Low",
    0x94 : "Resource Temporarily Not Available",
}
"""Dict of response code definitions. Missing values are reserved by ISO/SAE or proprietary."""

def translateRequestSID(sid : int) -> str:
    """Translates a UDS Request SID (service ID) to its service name."""
    return ServiceNames.get(sid, "Proprietary/Reserved")

def translateResponseSID(sid : int) -> str:
    """Translates a UDS Response SID (service ID) to its service name."""
    return ServiceNames.get(sid - 0x40, "Proprietary/Reserved")

def translateResponseCode(code : int) -> str:
    """Returns a description of the specified UDS response code."""
    # cover some specific cases
    if 0x95 <= code <= 0xEF:
        return "Reserved For Specific Conditions Not Correct"
    if 0xF0 <= code <= 0xFE:
        return "Conditions Not Correct: Vehicle Manufacturer Specific"
    return ResponseCodes.get(code, "ISO/SAE Reserved")

class UDSMessage:
    """
    Parent to UDS request & response subclasses.
    
    Each UDS message consists of 1 SID byte, with the existence of other properties
    depending on the service:
    - `sid`   (int - 1 byte)
    - `subfn` (int - 0 or 1 bytes)
    - `did`   (int - 0 or 2 bytes)
    - `data`  (bytes - n bytes)
    """

    _isResponse = False
    _sid    = None #type: int

    def __init__(self):
        self._hasSubfn   = False
        self._hasDID     = False
        self._hasData    = False
        self._dataSizeCanChange = False

        self._subfn  = None #type: int
        self._did    = None #type: int
        self._data   = b'' #type: bytes
        
        self._dataSize   = 0

    def name(self):
        if self._isResponse:
            return translateResponseSID(self._sid)
        else:
            return translateRequestSID(self._sid)

    def hasSubfn(self):
        return self._hasSubfn

    def hasDID(self):
        return self._hasDID

    def hasData(self):
        return self._hasData

    def dataSize(self):
        return self._dataSize

    def dataSizeCanChange(self):
        return self._dataSizeCanChange

    def isResponse(self):
        return self._isResponse

    def isRequest(self):
        return not self._isResponse

    def __str__(self) -> str:
        return self.raw.decode('utf-8', errors='replace')

    def __int__(self) -> int:
        return self.value # I'm torn between returning self.value and self._sid

    def __bytes__(self) -> bytes:
        return self.raw

    def __len__(self) -> int:
        return len(self.raw)

    def __getitem__(self, index : int) -> int:
        return self.raw[index]

    @classmethod
    def fromMessageData(cls, msg_data : bytes):
        """
        Generates a subclass instance of UDSMessage based on the given message data.

        Message data must be in the following format:
        1. Service ID (1 byte)
        2. Sub-Function ID (0 or 1 bytes)
        3. Data ID (0 or 2 bytes)
        4. Data (0 or n bytes)
        """
        msg = cls.fromSID(msg_data[0])
        index = 1
        if msg.hasSubfn():
            msg.subfn = msg_data[index]
            index += 1
        if msg.hasDID():
            msg.did = msg_data[index:index+2]
            index += 2
        if msg.hasData():
            if len(msg_data) > index:
                msg.data = msg_data[index:]
            else:
                msg.data = b''
        return msg

    @classmethod
    def fromSID(cls, sid : int):
        """
        Generates a subclass instance of UDSMessage based on the given message SID.

        Returns generic instance of this class if SID is not found in a subclass.
        """
        for msg in cls.__subclasses__():
            if msg._sid == sid:
                return msg()
        return cls()
            
    @property
    def sid(self) -> int:
        """
        Service ID (8-bit unsigned integer).

        This property is immutable.
        """
        return self._sid

    @property
    def subfn(self) -> int:
        """
        Sub-Function ID (8-bit unsigned integer).
        
        Will be None if the service does not have a sub-function.
        """
        if self._hasSubfn:
            return self._subfn
        else:
            return None

    @subfn.setter
    def subfn(self, val : int):
        if not self._hasSubfn: # attempted to assign subfn when sub-function doesn't exist for this service
            raise AttributeError("Attempted to set subfn of UDS service that doesn't have sub-functions available.")
        if not isinstance(val, int): # handle bytes & str
            val = int.from_bytes(sanitize_msg_param(val, 1), 'big')
        self._subfn = val & 0xFF

    @property
    def did(self) -> int:
        """
        Data ID (16-bit unsigned integer). Acts like a function argument in most cases.

        Will be None if the service does not have a DID.
        """
        if self._hasDID:
            return self._did
        else:
            return None

    @did.setter
    def did(self, val : int):
        if not self._hasDID: # attempted to assign DID when DID doesn't exist for this service
            raise AttributeError("Attempted to set DID of UDS service that doesn't include a DID.")
        if not isinstance(val, int): # handle bytes & str
            val = int.from_bytes(sanitize_msg_param(val, 2), 'big')
        self._did = val & 0xFFFF

    @property
    def data(self) -> bytes:
        """
        Data field (n bytes).

        Will be `b''` if the service does not have a data field.
        """
        if self._hasData:
            return self._data
        else:
            return b''

    @data.setter
    def data(self, val : bytes):
        if not self._hasData: # attempted to assign data when data doesn't exist for this service
            raise AttributeError("Attempted to set data of UDS service that doesn't include a data field.")
        if not isinstance(val, bytes):
            val = sanitize_msg_param(val)
        if len(val) != self._dataSize: # check for length mismatches
            if self._dataSizeCanChange: # update data size if allowed
                self._dataSize = len(val)
            elif len(val) < self._dataSize: # stuff with bytes to fill data size
                val += BYTE_STUFFING_VALUE * (self._dataSize - len(val)) # stuff to fit
            else:
                raise ValueError("Data value for UDS message is too large!")
        self._data = val

    @property
    def raw(self) -> bytes:
        """
        Raw bytes representing the service.
        
        Does not include protocol-specific data like CANID.

        Byteorder:
        1. Service ID (1 byte)
        2. Sub-Function ID (0 or 1 bytes)
        3. Data ID (0 or 2 bytes)
        4. Data (0 or n bytes)
        """
        val = b''
        val += sanitize_msg_param(self._sid, 1)
        if self._hasSubfn:
            val += sanitize_msg_param(self._subfn, 1)
        if self._hasDID:
            val += sanitize_msg_param(self._did, 2)
        if self._hasData:
            val += sanitize_msg_param(self._data, self._dataSize)
        return val

    @property
    def value(self) -> int:
        """
        Integer value of UDS message data field.

        If message doesn't contain a data field, returns 0.
        """
        if not self._data:
            return 0
        return int.from_bytes(self._data, 'big')

    @property
    def suppressPosRspMsgIndicationBit(self) -> bool:
        """
        Bit 7 of `subfn`.
        - True = positive response to this service is suppressed.
        - False = receiving ECU will send positive response.
        """
        if self._hasSubfn:
            return bool(self.subfn & 0b10000000)
        else:
            return False

    @suppressPosRspMsgIndicationBit.setter
    def suppressPosRspMsgIndicationBit(self, val : bool):
        if not self._hasSubfn:
            raise AttributeError("Attempted to set subfn of UDS service that doesn't have sub-functions available.")
        if val:
            self.subfn |= 0b10000000
        else:
            self.subfn &= 0b01111111



from . import UDSMessage


class TransferDataRequest(UDSMessage):
    """
    Transfer Data (Request)
    - `sid` = 0x36
    - `data` = blockSequenceCounter (1 byte) + transferEquestParameterRecord (n bytes)
    """

    _sid = 0x36
    _isResponse = False

    def __init__(self, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.data = data


class TransferDataResponse(UDSMessage):
    """
    Transfer Data (Response)
    - `sid` = 0x76
    - `data` = blockSequenceCounter (1 byte) + transferEquestParameterRecord (n bytes)
    """

    _sid = 0x76
    _isResponse = True

    def __init__(self, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.data = data



from . import UDSMessage
from .. import sanitize_msg_param


class CommunicationControlRequest(UDSMessage):
    """
    Communication Control (Request)
    - `sid` = 0x28
    - `subfn` = controlType
    - `comtype` = commnucationType (1 byte)
    - `high_node_id` = nodeIdentificationNumber (high byte) (1 byte)
    - `low_node_id` = nodeIdentificationNumber (low byte) (1 byte)
    """
    _sid = 0x28
    _isResponse = False

    # sub-function IDs. for convenience:
    enableRxAndTx = 0x00
    enableRxAndDisableTx = 0x01
    diableRxAndEnableTx = 0x02
    disableRxAndTx = 0x03
    enableRxAndDisableTxWithEnhancedAddressInformation = 0x04
    enableRxAndTxWithEnhancedAddressInformation = 0x05

    def __init__(self, subfn: int = enableRxAndTx, comtype: bytes = b'', high_node_id: bytes = b'', low_node_id: bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = True
        self._dataSizeCanChange = False

        self.subfn = subfn
        if self.subfn in {0x04, 0x05}:
            self._dataSize = len(comtype + high_node_id + low_node_id)
            self.data = comtype + high_node_id + low_node_id
        else:
            self._dataSize = len(comtype)
            self.data = comtype


class CommunicationControlResponse(UDSMessage):
    """
    Communication Control (Response)
    - `sid` = 0x68
    - `subfn` = controlType
    """
    _sid = 0x68
    _isResponse = True

    # sub-function IDs. for convenience:
    enableRxAndTx = 0x00
    enableRxAndDisableTx = 0x01
    diableRxAndEnableTx = 0x02
    disableRxAndTx = 0x03
    enableRxAndDisableTxWithEnhancedAddressInformation = 0x04
    enableRxAndTxWithEnhancedAddressInformation = 0x05

    def __init__(self, subfn: int = enableRxAndTx):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = False

        self.subfn = subfn



from . import UDSMessage
from .. import sanitize_msg_param


class WriteMemoryByAddressRequest(UDSMessage):
    """
    Write Memory By Address (Request)
    - `sid` = 0x3D
    - `did` = addressAndLengthFormatIdentifier (1 byte)
    - `data` = memoryAddress + memorySize + dataRecord
    """

    _sid = 0x3D
    _isResponse = False

    def __init__(self, did: int = 0, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = True
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.did = did
        self.data = data

    @property
    def did(self) -> int:
        return self._did

    @did.setter
    def did(self, val: int):
        """
        DID is 1 byte
        """
        val = int.from_bytes(sanitize_msg_param(val, 1), 'big')
        self._did = val & 0xFF


class WriteMemoryByAddressResponse(UDSMessage):
    """
    Write Memory By Address (Response)
    - `sid` = 0x7D
    - `did` = addressAndLengthFormatIdentifier (1 byte)
    - `data` = memoryAddress + memorySize
    """

    _sid = 0x7D
    _isResponse = True

    def __init__(self, did: int = 0, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = True
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.did = did
        self.data = data

    @property
    def did(self) -> int:
        return self._did

    @did.setter
    def did(self, val: int):
        """
        DID is 1 byte
        """
        val = int.from_bytes(sanitize_msg_param(val, 1), 'big')
        self._did = val & 0xFF



"""A package for parsing and creating Unified Diagnostic Services messages."""

from . import *
from .UDS import *
from .DiagnosticSessionControl import *
from .ECUReset import *
from .WriteDataByIdentifier import *
from .ReadDataByIdentifier import *
from .RequestDownload import *
from .SecurityAccess import *
from .Authentication import *
from .CommunicationControl import *
from .ControlDTCSetting import *
from .LinkControl import *
from .ReadMemoryByAddress import *
from .ReadScalingDataByIdentifier import *
from .ResponseOnEvent import *
from .TesterPresent import *
from .ClearDiagnosticInformation import *
from .InputOutputControlByIdentifier import *
from .ReadDTCInformation import *
from .RequestFileTransfer import *
from .RequestTransferExit import *
from .RequestUpload import *
from .RoutineControl import *
from .TransferData import *
from .WriteMemoryByAddress import *
from .SecuredDataTransmission import *
from .ReadDataByPeriodicIdentifier import *
from .DynamicallyDefineDataIdentifier import *


from . import UDSMessage


class ReadDataByIdentifierRequest(UDSMessage):
    """
    Read Data By Identifier (Request)
    - `sid` = 0x22
    - `did` = dataIdentifier
    """

    _sid = 0x22
    _isResponse = False

    def __init__(self, did: int = 0x00):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = True
        self._hasData = False

        self.did = did


class ReadDataByIdentifierResponse(UDSMessage):
    """
    Read Data By Identifier (Response)
    - `sid` = 0x62
    - `did` = dataIdentifier
    - `data` = dataRecord (n bytes)
    """

    _sid = 0x62
    _isResponse = True

    def __init__(self, did: int = 0x00, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = True
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.did = did
        self.data = data



from . import UDSMessage

class WriteDataByIdentifierRequest(UDSMessage):
    """
    Write Data By Identifier (Request)
    - `sid` = 0x2E
    - `did` = dataIdentifier
    - `data` = dataRecord (n bytes)
    """

    _sid = 0x2E
    _isResponse = False

    def __init__(self, did : int = 0, data : bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = True
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.did = did
        self.data = data

    @property
    def dataIdentifier(self) -> int:
        """`dataIdentifier` field, stored in `did`."""
        return self.did

    @dataIdentifier.setter
    def dataIdentifier(self, value : int):
        self.did = value

    @property
    def dataRecord(self) -> bytes:
        """`dataRecord` field, stored in `data`."""
        return self.data

    @dataRecord.setter
    def dataRecord(self, value : bytes):
        self.data = value

class WriteDataByIdentifierResponse(UDSMessage):
    """
    Write Data By Identifier (Response)
    - `sid` = 0x6E
    - `did` = dataIdentifier
    """

    _sid = 0x6E
    _isResponse = True

    def __init__(self, did : int = 0):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = True
        self._hasData = False

        self.did = did

    @property
    def dataIdentifier(self) -> int:
        """`dataIdentifier` field, stored in `did`."""
        return self.did

    @dataIdentifier.setter
    def dataIdentifier(self, value : int):
        self.did = value

    # @property
    # def dataRecord(self) -> bytes:
    #     """`dataRecord` field, stored in `data`."""
    #     return self.data

    # @dataRecord.setter
    # def dataRecord(self, value : bytes):
    #     self.data = value



from . import UDSMessage


class TesterPresentRequest(UDSMessage):
    """
    TesterPresent (Request)
    - `sid` = 0x3E
    - `subfn` = zeroSubFunction
    """

    _sid = 0x3E
    _isResponse = False

    # sub-function IDs, for convenience:
    zeroSubFunction = 0x00

    def __init__(self, subfn: int = zeroSubFunction):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = False

        self.subfn = subfn


class TesterPresentResponse(UDSMessage):
    """
    Tester Present (Response)
    - `sid` = 0x7E
    - `subfn` = zeroSubFunction
    """

    _sid = 0x7E
    _isResponse = True

    zeroSubFunction = 0x00

    def __init__(self, subfn: int = zeroSubFunction):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = False

        self.subfn = subfn



from . import UDSMessage


class AuthenticationRequest(UDSMessage):
    """
    Authentication (Request)
    - `sid` = 0x29
    - `subfn` = authenticationTask 
    """

    _sid = 0x29
    _isResponse = False

    # sub-function IDs, for convenience:
    deAuthenticate = 0x00
    authenticationConfiguration = 0x08
    # only if authentication with PKI Certificate Exchange (APCE) is used
    verifyCertificateUnidirectional = 0x01
    verifyCertificateBidirectional = 0x02
    proofOfOwnership = 0x03
    transmitCertificate = 0x04
    # only if authentication with Challenge-Response (ACR) is used
    requestChallengeForAuthentication = 0x05
    verifyProofOfOwnershipUnidirectional = 0x06
    verifyProofOfOwnershupBidirectional = 0x07
    authenticationConfiguration = 0x08

    def __init__(self, subfn: int = authenticationConfiguration):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = False

        self.subfn = subfn


class AuthenticationResponse(UDSMessage):
    """
    Authentication (Response)
    - `sid` = 0x69
    - `subfn` = authenticationTask
    - `data` = authenticationReturnParameter + other parameters depends on subfn
    """

    _sid = 0x69
    _isResponse = True

    # sub-function IDs, for convenience:
    deAuthenticate = 0x00
    authenticationConfiguration = 0x08
    # only if authentication with PKI Certificate Exchange (APCE) is used
    verifyCertificateUnidirectional = 0x01
    verifyCertificateBidirectional = 0x02
    proofOfOwnership = 0x03
    transmitCertificate = 0x04
    # only if authentication with Challenge-Response (ACR) is used
    requestChallengeForAuthentication = 0x05
    verifyProofOfOwnershipUnidirectional = 0x06
    verifyProofOfOwnershupBidirectional = 0x07
    authenticationConfiguration = 0x08

    def __init__(self, subfn: int = authenticationConfiguration, data: bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.subfn = subfn
        self.data = data



from . import UDSMessage


class ReadDTCInformationRequest(UDSMessage):
    """
    Read DTC information (Request)
    - `sid` = 0x19
    - `subfn` = reportType
    - `data` => depends on subfn
    """

    _sid = 0x19
    _isResponse = False

    # sub-function IDs, for convenience:
    reportNumberOfDTCByStatusMask = 0x01
    reportDTCByStatusMask = 0x02
    reportDTCSnapshotIdentification = 0x03
    reportDTCSnapshotRecordByDTCNumber = 0x04
    reportDTCStoredDataByRecordNumber = 0x05
    reportDTCExtDataRecordByDTCNumber = 0x06
    reportDTCNumberOfDTCBySecerityMaskRecord = 0x07
    reportDTCBySeverityMaskRecord = 0x08
    reportSeverityInformationOfDTC = 0x09
    reportSupportedDTC = 0x0A
    reposrtFirstTestFailedDTC = 0x0B
    reportFirstConfirmedDTC = 0x0C
    reportMostRecentTestFailedDTC = 0x0D
    reportMostRecentConfirmedDTC = 0x0E
    reportDTCFaultDetectionCounter = 0x14
    reportDTCWithPermanentStatus = 0x15
    reportDTCExtDataRecordByRecordNumber = 0x16
    reportUserDefMemoryDTCByStatusMask = 0x17
    reportUserDefMemoryDTCSnapshotRecordByDTCNumber = 0x18
    reportUserDefMemoryDTCExtDataRecordByDTCNumber = 0x19
    reportDTCExtendedDataRecordIdentification = 0x1A
    reportWWHOBDDTCByMaskRecord = 0x42
    reportWWHOBDDTCWithPermanentStatus = 0x55
    reportDTCInformationByDTCReadinessGroupIdentifer = 0x56

    def __init__(self, subfn: int = reportNumberOfDTCByStatusMask, data: bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.subfn = subfn
        self.data = data


class ReadDTCInformationResponse(UDSMessage):
    """
    Read DTC Information (Response)
    - `sid` = 0x59
    - `subfn` = reportType
    - `data` => depends on subfn
    """

    _sid = 0x59
    _isResponse = True

    # sub-function IDs, for convenience:
    reportNumberOfDTCByStatusMask = 0x01
    reportDTCByStatusMask = 0x02
    reportDTCSnapshotIdentification = 0x03
    reportDTCSnapshotRecordByDTCNumber = 0x04
    reportDTCStoredDataByRecordNumber = 0x05
    reportDTCExtDataRecordByDTCNumber = 0x06
    reportDTCNumberOfDTCBySecerityMaskRecord = 0x07
    reportDTCBySeverityMaskRecord = 0x08
    reportSeverityInformationOfDTC = 0x09
    reportSupportedDTC = 0x0A
    reposrtFirstTestFailedDTC = 0x0B
    reportFirstConfirmedDTC = 0x0C
    reportMostRecentTestFailedDTC = 0x0D
    reportMostRecentConfirmedDTC = 0x0E
    reportDTCFaultDetectionCounter = 0x14
    reportDTCWithPermanentStatus = 0x15
    reportDTCExtDataRecordByRecordNumber = 0x16
    reportUserDefMemoryDTCByStatusMask = 0x17
    reportUserDefMemoryDTCSnapshotRecordByDTCNumber = 0x18
    reportUserDefMemoryDTCExtDataRecordByDTCNumber = 0x19
    reportDTCExtendedDataRecordIdentification = 0x1A
    reportWWHOBDDTCByMaskRecord = 0x42
    reportWWHOBDDTCWithPermanentStatus = 0x55
    reportDTCInformationByDTCReadinessGroupIdentifer = 0x56

    def __init__(self, subfn: int = reportNumberOfDTCByStatusMask, data: bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.subfn = subfn
        self.data = data



from typing import Union
from . import UDSMessage
from .. import sanitize_msg_param

class RequestUploadRequest(UDSMessage):
    """
    Request Upload (Request)
    - `sid` = 0x35
    - `data` = dfid + alfid + maddr + msize
        - `dfid' = dataFormatIdentifier (1 byte)
        - `alfid' = addressAndLengthFormatIdentifier (1 byte)
            - bits 7-4: define length (number of bytes) of memorySize
            - bits 3-0`: define length (number of bytes) of memoryAddress
        - `maddr` = memoryAddress (n bytes)
        - `msize` = memorySize (n bytes)
    """

    _sid = 0x35
    _isResponse = False

    def __init__(self, dfid: bytes = b'\x00', alfid: bytes = b'\x00', maddr: bytes = b'', msize: bytes = b'', autoALFID: bool = True):
        """
        - `dfid`: Data Format Identifier (1 byte). Default is `0x00` (no encryption or compression), other values are manufacturer-specific.
        - `alfid`: Address and Length Format Identifer (1 byte). Can be left default if `autoALFID` is `True`.
            - bits 7-4: define length (number of bytes) of msize
            - bits 3-0: define length (number of bytes) of maddr
        - `maddr`: memoryAddress (0-15 bytes)
        - `msize`: memorySize (0-15 bytes)
        - `autoALFID`: If `True`, `alfid` is automatically calculated based on the values of `maddr` and `msize`. If `False`, `alfid` must
        be specified manually to produce a valid request.
        """
        super().__init__()
        self._hasSubfn:          bool = False
        self._hasDID:            bool = False
        self._hasData:           bool = True
        self._dataSizeCanChange: bool = True

        self._autoALFID: bool = autoALFID
        self.dfid             = dfid
        self.alfid            = alfid
        self.maddr            = maddr
        self.msize            = msize

        self._dataSize: int = len(self.data)

    @property
    def dfid(self) -> int:
        return int.from_bytes(self._dfid, 'big')

    @dfid.setter
    def dfid(self, dfid: Union[int, bytes]) -> None:
        self._dfid = sanitize_msg_param(dfid, 1)
    
    @property
    def alfid(self) -> int:
        return int.from_bytes(self._alfid, 'big')

    @alfid.setter
    def alfid(self, alfid: Union[int, bytes]) -> None:
        self._alfid = sanitize_msg_param(alfid, 1)

    @property
    def maddr(self) -> int:
        return int.from_bytes(self._maddr, 'big')

    @maddr.setter
    def maddr(self, maddr: Union[int, bytes]) -> None:
        maddr = sanitize_msg_param(maddr)
        if self._autoALFID: # Auto-calculating alfid
            if len(maddr) > 0x0f:
                raise ValueError(F"maddr length must be less than 16 bytes, got {len(maddr)} bytes: {maddr}")
            self.alfid = (self.alfid & 0xf0) | (len(maddr) & 0x0f)
        self._maddr = maddr

    @property
    def msize(self) -> int:
        return int.from_bytes(self._msize, 'big')

    @msize.setter
    def msize(self, msize: Union[int, bytes]) -> None:
        msize = sanitize_msg_param(msize)
        if self._autoALFID: # Auto-calculating alfid
            if len(msize) > 0x0f:
                raise ValueError(F"msize length must be less than 16 bytes, got {len(msize)} bytes: {msize}")
            self.alfid = (self.alfid & 0x0f) | ((len(msize) & 0x0f) << 4)
        self._msize = msize

    @property
    def data(self) -> bytes:
        """
        Concatenates the request's properties to form the data field.
        """
        return sanitize_msg_param(self._dfid) + sanitize_msg_param(self._alfid) + sanitize_msg_param(self._maddr) + sanitize_msg_param(self._msize)


class RequestUploadResponse(UDSMessage):
    """
    Request Upload (Response)
    - `sid` = 0x75
    - `data` = lfid + mnrob
        - `lfid` = lengthFormatIdentifier (1 byte)
            - bit 7 - 4: maxNumberOfBlockLength
            - bit 3 - 0: reserved; set to be 0
        - `mnrob` = maxNumberOfBlockLength (n bytes)
    """

    _sid = 0x75
    _isResponse = True

    def __init__(self, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.data = data

    @property
    def lfid(self) -> int:
        """
        Length format identifier. Only bits 7-4 are used.
        """
        return self.data[0]

    @lfid.setter
    def lfid(self, lfid: Union[int, bytes]) -> None:
        self.data = sanitize_msg_param(lfid, 1) + self.data[1:]

    @property
    def mnrob(self) -> int:
        """
        Max NumbeR Of Blocks. Length must match the value specified in the LFID.
        """
        return int.from_bytes(self.data[1:], 'big')

    @mnrob.setter
    def mnrob(self, mnrob: Union[int, bytes]) -> None:
        self.data = sanitize_msg_param(self.data[0]) + sanitize_msg_param(mnrob)



from . import UDSMessage


class InputOutputControlByIdentifierRequest(UDSMessage):
    """
    Input Output Control By Identifier (Request)
    - `sid` = 0x2F
    - `did` = dataIdentifier
    - `data` = controlOptionRecord + controlEnableMaskRecord
    """

    _sid = 0x2F
    _isResponse = False

    def __init__(self, did: int = 0, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = True
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.did = did
        self.data = data

class InputOutputControlByIdentifierResponse(UDSMessage):
    """
    Input Output Control By Identifier (Response)
    - `sid` = 0x6F
    - `did` = dataIdentifier
    - `data` = controlStatusRecord
    """

    _sid=0x6F
    _isResponse=True

    def __init__(self, did: int = 0, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = True
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.did = did
        self.data = data
        


from . import UDSMessage


class RoutineControlRequest(UDSMessage):
    """
    Routine Control (Request)
    - `sid` = 0x31
    - `subfn` = routineControlType
    - 'did` = routineIdentifier
    - `data` = routineControlOptionRecord
    """

    _sid = 0x31
    _isResponse = False

    # sub-function IDs, for convenience:
    startRoutine = 0x01
    stopRoutine = 0x02
    requestRoutineResults = 0x03

    def __init__(self, subfn: int = startRoutine, did: int = 0, data: bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = True
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.subfn = subfn
        self.did = did
        self.data = data


class RoutineControlResponse(UDSMessage):
    """
    Routine Control (Response)
    - `sid` = 0x71
    - `subfn` = routineControlType
    - 'did` = routineIdentifier
    - `data` = routineInfo (1 byte) + routineStatusRecord (n bytes)
    """

    _sid = 0x71
    _isResponse = True

    # sub-function IDs, for convenience:
    startRoutine = 0x01
    stopRoutine = 0x02
    requestRoutineResults = 0x03

    def __init__(self, subfn: int = startRoutine, did: int = 0, data: bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = True
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.subfn = subfn
        self.did = did
        self.data = data



from . import UDSMessage


class SecuredDataTransmissionRequest(UDSMessage):
    """
    Secured Data Transmission (Request)
    - `sid` = 0x84
    - `data` =  Administrative Parameter (2 bytes) +
                Signature/Encryption Calculation (1 byte) +
                Signature Length (2 bytes) +
                Anti-replay Counter (2 bytes) +
                Internal Message Service Request ID (1 byte) +
                Service Specific Parameter (n bytes) +
                Signature/MAC (n bytes)
    """

    _sid = 0x84
    _isResponse = False

    def __init__(self, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.data = data


class SecuredDataTransmissionResponse(UDSMessage):
    """
    Secured Data Transmission (Response)
    - `sid` = 0xC4
    - `data` =  Administrative Parameter (2 bytes) +
                Signature/Encryption Calculation (1 byte) +
                Signature Length (2 bytes) +
                Anti-replay Counter (2 bytes) +
                Internal Message Service Response ID (1 byte) +
                Response Specific Parameter (n bytes) +
                Signature/MAC (n bytes)
    """

    _sid = 0xC4
    _isResponse = True

    def __init__(self, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.data = data



from . import UDSMessage
from .. import sanitize_msg_param

class ECUResetRequest(UDSMessage):
    """
    ECU Reset (Request)
    - `sid` = 0x11
    - `subfn` = resetType
    """

    _sid = 0x11
    _isResponse = False

    # sub-function IDs, for convenience:
    hardReset = 0x01
    keyOffOnReset = 0x02
    softReset = 0x03
    enableRapidPowerShutDown = 0x04
    disableRapidPowerShutDown = 0x05
    
    def __init__(self, subfn : int = hardReset):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = False
        self.subfn = subfn

class ECUResetResponse(UDSMessage):
    """
    ECU Reset (Response)
    - `sid` = 0x51
    - `subfn` = resetType
    - `data` = powerDownTime (0 or 1 bytes)
       1. powerDownTime (only present for subfn = 0x04)
    """

    _sid = 0x51
    _isResponse = True

    # sub-function IDs, for convenience:
    hardReset = 0x01
    keyOffOnReset = 0x02
    softReset = 0x03
    enableRapidPowerShutDown = 0x04
    disableRapidPowerShutDown = 0x05
    
    def __init__(self, subfn : int = hardReset, data = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._dataSizeCanChange = False
        self.subfn = subfn

        if self.subfn == self.enableRapidPowerShutDown or data:
            self._hasData = True
            self._dataSize = 1
            self.data = sanitize_msg_param(data, 1)
        else:
            self._hasData = False

    # overload subfn, since changing subfn will change data
    @property
    def subfn(self) -> int:
        return self._subfn

    @subfn.setter
    def subfn(self, val : int):
        if not isinstance(val, int): # handle bytes & str
            val = int.from_bytes(sanitize_msg_param(val, 1), 'big')
        self._subfn = val & 0xFF
        if self._subfn == 0x04: # subfn 0x04 has 1 data byte
            self._hasData = True
            self._dataSize = 1
            self.data = b'\xFF'

    @property
    def powerDownTime(self) -> int:
        """
        `powerDownTime` value in message data. This value should only exist for SID = 0x04.

        Will return None if not available.
        """
        if not self._hasData:
            return None
        return self.data[0]

    @powerDownTime.setter
    def powerDownTime(self, val : int):
        self.data = sanitize_msg_param(val, 1)



from . import UDSMessage


class LinkControlRequest(UDSMessage):
    """
    Link Control (Request)
    - `sid` = 0x87
    - `subfn` = linkControlType
    - `data` = (1 or 0 btye)
        1. linkControlModelIdentifier (for verifyModeTransitionWithFixedParameter)
        2. linkRecord (for verifyModeTransitionWithSpecificParameter)
        3. None (for transitionMode)
    """

    _sid = 0x87
    _isResponse = False

    # sub-function IDs, for convenience:
    verifyModeTransitionWithFixedParameter = 0x00
    verifyModeTransitionWithSpecificParameter = 0x02
    transitionMode = 0x03

    def __init__(self, subfn: int = verifyModeTransitionWithFixedParameter, data: bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False

        self.subfn = subfn
        if subfn != self.transitionMode:
            self._hasData = True
            self._dataSize = len(data)
            self._dataSizeCanChange = False
            self.data = data
        else:
            self._hasData = False


class LinkControlResponse(UDSMessage):
    """
    Link Control (Response)
    - `sid` = 0xC7
    - `subfn` = linkControlType
    """

    _sid = 0xC7
    _isResponse = True

    # sub-function IDs, for convenience:
    verifyModeTransitionWithFixedParameter = 0x00
    verifyModeTransitionWithSpecificParameter = 0x02
    transitionMode = 0x03

    def __init__(self, subfn: int = verifyModeTransitionWithFixedParameter):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = False

        self.subfn = subfn



from . import UDSMessage


class ReadDataByPeriodicIdentifierRequest(UDSMessage):
    """
    Read Data By Periodic Identifier (Request)
    - `sid` = 0x2A
    - `data` = transmissionMode (1 byte) + periodicDataIdentifier (n bytes)
    """

    _sid = 0x2A
    _isResponse = False

    def __init__(self, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.data = data


class ReadDataByPeriodicIdentifierResponse(UDSMessage):
    """
    Read Data By Periodic Identifier (Response)
    - `sid` = 0x6A
    - `data` = transmissionMode (1 byte) + periodicDataIdentifier (n bytes)
    """

    _sid = 0x6A
    _isResponse = True

    def __init__(self):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = False
        self._hasData = False



from . import UDSMessage


class DynamicallyDefineDataIdentifierRequest(UDSMessage):
    """
    Dynamically Define Data Identifier (Request)
    - `sid` = 0x2C
    - `subfn` = definitionType
    - `did` = dynamicallyDefinedDataIdentifier = byte #1 (MSB F2/F3) + byte #2 (LSB 00 - FF)
    - `data` => depends on subfn
    """

    _sid = 0x2C
    _isResponse = False

    # sub-function IDs, for convenience:
    defineByIdentifier = 0x01
    defineByMemoryAddress = 0x02
    clearDynamicallyDefineddataIdentifier = 0x03

    def __init__(self, subfn: int = defineByIdentifier, did: int = 0, data: bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = True

        if subfn != self.clearDynamicallyDefineddataIdentifier:
            self._hasData = True
            self._dataSize = len(data)
            self._dataSizeCanChange = True
            self.data = data
        else:
            self._hasData = False

        self.subfn = subfn
        self.did = did


class DynamicallyDefineDataIdentifierResponse(UDSMessage):
    """
    Dynamically Define Data Identifier (Response)
    - `sid` = 0x6C
    - `subfn` = definitionType
    - `did` = dynamicallyDefinedDataIdentifier = byte #1 (MSB F2/F3) + byte #2 (LSB 00 - FF)
    """

    _sid = 0x6C
    _isResponse = True

    # sub-function IDs, for convenience:
    defineByIdentifier = 0x01
    defineByMemoryAddress = 0x02
    clearDynamicallyDefineddataIdentifier = 0x03

    def __init__(self, subfn: int = defineByIdentifier, did: int = 0):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = True
        self._hasData = False

        self.subfn = subfn
        self.did = did



from . import UDSMessage

class DiagnosticSessionControlRequest(UDSMessage):
    """
    Diagnostic Session Control (Request)
    - `sid` = 0x10
    - `subfn` = diagnosticSessionType
    """
    _sid = 0x10
    _isResponse = False

    # sub-function IDs, for convenience:
    defaultSession = 0x01
    programmingSession = 0x02
    extendedDiagnosticSession = 0x03
    safetyDiagnosticSession = 0x04

    def __init__(self, subfn : int = defaultSession):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = False
        self.subfn = subfn

class DiagnosticSessionControlResponse(UDSMessage):
    """
    Diagnostic Session Control (Response)
    - `sid` = 0x50
    - `subfn` = diagnosticSessionType
    - `data` = sessionParameterRecord (4 bytes)
       1. P2Server_Max (high)
       2. P2Server_Max (low)
       3. P2*Server_Max (high)
       4. P2*Server_Max (low)
    """

    _sid = 0x50
    _isResponse = True

    # sub-function IDs, for convenience:
    defaultSession = 0x01
    programmingSession = 0x02
    extendedDiagnosticSession = 0x03
    safetyDiagnosticSession = 0x04

    def __init__(self, subfn : int = defaultSession, data : bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = True
        self._dataSize = 4
        self._dataSizeCanChange = False
        self.subfn = subfn
        self.data = data



from . import UDSMessage


class SecurityAccessRequest(UDSMessage):
    """`
    Security Access (Request)
    - `sid` = 0x27
    - `subfn` = securityAccesType
    - `data` = securityAccesDataRecord (for requestSeed) 
                or securityKey (for sendKey) (n btyes)
    """

    _sid = 0x27
    _isResponse = False

    # sub-function IDs (bits 6 - 0), for convenience
    # with the level of security defined by the vehicle manufacturer
    requestSeed = 0x01
    sendKey = 0x02
    # with different levels of security defined by the vehicle manufacturer
    # requestSeed = 0x03, 0x05, 0x07 to 0x41
    # sendKey = 0x04, 0x06, 0x08 to 0x42

    def __init__(self, subfn: int = 0x01, data: bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.subfn = subfn
        self.data = data


class SecurityAccessResponse(UDSMessage):
    """
    Security Access (Response)
    - `sid` = 0x67
    - `subfn` = securityAccesType
    - `data` = securitySeed (n bytes)
    """

    _sid = 0x67
    _isResponse = True

    # sub-function IDs (bits 6 - 0), for convenience
    requestSeed = 0x01
    # with different levels of security defined by the vehicle manufacturer
    # requestSeed = 0x03, 0x05, 0x07 to 0x41

    def __init__(self, subfn: int = 0x01, data: bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.subfn = subfn
        self.data = data



from . import UDSMessage


class ControlDTCSettingRequest(UDSMessage):
    """
    Control DTC Setting (Request)
    - `sid` = 0x85
    - `subfn` = DTCSettingType
    - `data` = DTCSettingControlOptionRecord
    """

    _sid = 0x85
    _isResponse = False

    # sub-function IDs, for convenience:
    on = 0x01
    off = 0x02

    def __init__(self, subfn: int = on, data: bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.subfn = subfn
        self.data = data


class ControlDTCSettingResponse(UDSMessage):
    """
    Control DTC Setting (Response)
    - `sid` = 0xC5
    - `subfn` = DTCSettingType
    """

    _sid = 0xC5
    _isResponse = True

    # sub-function IDs, for convenience:
    on = 0x01
    off = 0x02   

    def __init__(self, subfn: int = on):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = False

        self.subfn = subfn



from . import UDSMessage


class RequestFileTransferRequest(UDSMessage):
    """
    Request File Transfer (Request)
    - `sid` = 0x38
    - `data` =  modeOfOperation (1 byte) +
                filePathAndNameLength (2 bytes) +
                filePathAndName (n bytes) +
                dataFormatIdentifier (0 - 1 byte) +
                fileSizeParameterLength (0 - 1 byte) +
                fileSizeUnCompressed (0 - n bytes) +
                fileSizeCompressed (0 - n bytes)
    """

    _sid = 0x38
    _isResponse = False

    # modeOfOperation:
    AddFile = 0x01
    DeleteFile = 0x02
    ReplaceFile = 0x03
    ReadFile = 0x04
    ReadDir = 0x05
    ResumeFile = 0x06

    def __init__(self, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.data = data


class RequestFileTransferResponse(UDSMessage):
    """
    Request File Transfer (Response)
    - `sid` = 0x78
    - `data` =  modeOfOperation (1 byte) +
                lengthFormatIdentifier (0 - 1 byte) +
                maxNumberOfBlockLength (0 - n bytes) +
                dataFormatIdentifier (0 - 1 byte) +
                fileSizeOrDirInfoParameterLength (0 or 2 bytes) +
                fileSizeUpcompressedOrDirInfoLength (0 - n bytes) +
                fileSizeCompressed (0 - n bytes) +
                filePosition (0 - 8 bytes)
    """

    _sid = 0x78
    _isResponse = True

    # modeOfOperation:
    AddFile = 0x01
    DeleteFile = 0x02
    ReplaceFile = 0x03
    ReadFile = 0x04
    ReadDir = 0x05
    ResumeFile = 0x06

    def __init__(self, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.data = data



from . import UDSMessage


class ResponseOnEventRequest(UDSMessage):
    """
    Response On Event (Request)
    - `sid` = 0x86
    - `subfn` = eventType
    - `data` = eventTypeRecord + serviceToRespondToRecord
        - `eventTypeRecord` = eventTypeParameter
        - `serviceToRespondToRecord` = serviceId + serviceParameter
            (serviceId is mandatory to be present if the SubFunction parameter is not equal t0 RepostActivatedEvents, 
            stopResponseOnEvent, startResponseOnEvent, clearResponseOnEvent)
    """

    _sid = 0x86
    _isResponse = False

    # sub-function = storageState (bit 6) + parameter (bit 5 to 0)
    # storageState
    doNotStoreEvent = 0x00
    storeEvent = 0x01

    # parameter
    stopResponseOnEvent = 0x00
    onDTCStatusChange = 0x01
    onChangeOfDataIdentifier = 0x03
    reportActivatedEvents = 0x04
    startResponseOnEvent = 0x05
    clearResponseOnEvent = 0x06
    onComparisonOfValues = 0x07
    reportMostRecentDtcOnStatusChange = 0x08
    reportDTCRecordInformationOnDtcStatusChange = 0x09

    def __init__(self, subfn: int = 0x00, data: bytes = b''):
        super().__init__()
        self._hasSubfn = True
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.subfn = subfn
        self.data = data


class ResponseOnEventResponse(UDSMessage):
    """
    Response On Event (Response)
    - `sid` = 0xC6
    - `did` = eventType + numberOfIdentifiedEvents + eventWindowTime
    - `data` = eventTypeRecord + serviceToRespondToRecord + other data depends on subfn
        - `eventTypeRecord` = eventTypeParameter
        - `serviceToRespondToRecord` = serviceId + serviceParameter
    """

    _sid = 0xC6
    _isResponse = True

    def __init__(self, did: int = 0, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = True
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.did = did
        self.data = data



from . import UDSMessage


class ReadMemoryByAddressRequest(UDSMessage):
    """
    Read Memory By Address (Request)
    - `sid` = 0x23
    - `did` = addressAndLengthFormatIdentifier
    - `data` = memoryAddress + memorySize
    """

    _sid = 0x23
    _isResponse = False

    def __init__(self, did: int = 0, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = True
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.did = did
        self.data = data


class ReadMemoryByAddressResponse(UDSMessage):
    """
    Read Memory By Address (Response)
    - `sid` = 0x63
    - `data` = dataRecord
    """

    _sid = 0x63
    _isResponse = True

    def __init__(self, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.data = data



from . import UDSMessage


class ClearDiagnosticInformationRequest(UDSMessage):
    """
    Clear Diagnostic Information (Request)
    - `sid` = 0x14
    - `data` = groupOfDTC + Memory Selection
    """

    _sid = 0x14
    _isResponse = False

    def __init__(self, data: bytes = b''):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = False
        self._hasData = True
        self._dataSize = len(data)
        self._dataSizeCanChange = True

        self.data = data


class ClearDiagnosticInformationResponse(UDSMessage):
    """
    Clear Diagnostic Information (Response)
    - `sid` = 0x54
    """

    _sid = 0x54
    _isResponse = True

    def __init__(self):
        super().__init__()
        self._hasSubfn = False
        self._hasDID = False
        self._hasData = False
